import yaml
import os
import stat
import boto3

from datetime import datetime
from functools import partial

from cement import Controller, ex
from cement.utils import fs
from cement.utils import shell

from ..core.utils import environment_names, job_names, find_nearest_root
from ..core.environments.stack_collection import StackCollection
from ..core.clusters.cluster_booter import ClusterBooter
from ..core.clusters.cluster_instance import ClusterInstance
from .arguments import profile as profile_arg
from .arguments import cluster_id as cluster_id_arg
from .arguments import ssh_key_location as ssh_key_location_arg


class Clusters(Controller):

    class Meta:
        label = 'clusters'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'create and manage EMR clusters'

    @ex(
        help='list clusters',
        arguments=[
            (['-a', '--all'], {
                'help': 'Filter list to all the clusters intead of just ones you created',
                'action': 'store_true'
            }),
            (['-s', '--state'], {
                'help': 'By default we only show WAITING, RUNNING clusters pass in -s multiple times for multiple states',
                'action': 'append',
                'choices': ['STARTING', 'BOOTSTRAPPING', 'RUNNING', 'WAITING', 'TERMINATING', 'TERMINATED', 'TERMINATED_WITH_ERRORS'],
                'default': ['WAITING', 'RUNNING']
            }),
            profile_arg()
        ]
    )
    def list(self):
        profile = self.app.pargs.profile
        states = self.app.pargs.state
        allclusters = self.app.pargs.all

        session = boto3.session.Session(profile_name=profile)
        client = session.client('emr')
        paginator = client.get_paginator('list_clusters')

        table = self.app.db.table('users')
        user_id = table.all()[0]['id']

        def mine(cluster):
            for tag in cluster['Tags']:
                if tag['Key'] == 'user_id' and tag['Value'] == user_id:
                    return True
            return False

        clusters = []
        for response in paginator.paginate(ClusterStates=states):
            clusters = [*response['Clusters'], *clusters]

        clusters = [client.describe_cluster(ClusterId=cluster['Id'])[
            'Cluster'] for cluster in clusters]

        if not allclusters:
            clusters = [cluster for cluster in clusters if mine(cluster)]

        for cluster in clusters:
            self.app.log.info(yaml.dump(cluster))

    @ex(
        help='Install Dependencies from config file',
        arguments=[
            cluster_id_arg(),
            ssh_key_location_arg(),
            profile_arg(),
            (['-f', '--config-file'], {
                'help': 'Path to the cluster config file (example: ./jobs/config/clusters/dev/config.yml)',
                'required': True
            })
        ]
    )
    def install(self):

        cluster_id = self.app.pargs.cluster_id
        profile = self.app.pargs.profile
        ssh_key = self.app.pargs.ssh_key
        config_file = self.app.pargs.config_file
        log = self.app.log

        session = boto3.session.Session(profile_name=profile)
        client = session.client('emr')
        response = client.describe_cluster(ClusterId=cluster_id)

        paginator = client.get_paginator('list_instances')
        response_iterator = paginator.paginate(ClusterId=cluster_id)

        # try:
        cluster_conf = open(config_file, "r").read()
        pips = yaml.load(cluster_conf).get('PipInstall', [])
        # except:
        #    log.error("Sorry we couldn't find anything to pip install in your config file")
        #    return

        for response in response_iterator:
            for instance in response['Instances']:
                instance = ClusterInstance(
                    instance, {"key_filename": [ssh_key]})
                log.info("\nInstalling dependencies on %s" %
                         instance.public_dns_name)
                resp = instance.run_cmd(
                    "sudo /usr/bin/python3 -m pip install %s" % (" ".join(pips)))
                log.info(resp)

    @ex(
        help='Run a command on the remote cluster',
        arguments=[
            (['-r', '--run-cmd'], {
                'help': 'A string representing the command you want to run',
                'required': True
            }),
            (['-m', '--master-only'], {
                'help': 'Run only on master',
                'action': 'store_true'
            }),
            profile_arg(),
            cluster_id_arg(),
            ssh_key_location_arg()
        ]
    )
    def run_cmd(self):
        cluster_id = self.app.pargs.cluster_id
        profile = self.app.pargs.profile
        master_only = self.app.pargs.master_only
        ssh_key = self.app.pargs.ssh_key
        log = self.app.log

        session = boto3.session.Session(profile_name=profile)
        client = session.client('emr')
        response = client.describe_cluster(ClusterId=cluster_id)
        master_dns = response['Cluster']['MasterPublicDnsName']

        paginator = client.get_paginator('list_instances')
        response_iterator = paginator.paginate(ClusterId=cluster_id)
        instances = []
        for response in response_iterator:
            for instance in response['Instances']:
                instances.append(ClusterInstance(
                    instance, {"key_filename": [ssh_key]}))

        if master_only:
            instances = [
                instance for instance in instances if instance.public_dns_name == master_dns]

        for instance in instances:
            log.info("\nRunning command on host %s" % instance.public_dns_name)
            resp = instance.run_cmd(self.app.pargs.run_cmd)
            log.info(resp)

    @ex(
        help='Copy files from local machine to every node in the cluster',
        arguments=[
            (['-c', '--cluster-id'], {
                'help': 'Id of the cluster you want to copy files to (hint: use list to show clusters)',
                'action': 'store'
            }),
            (['-s', '--source'], {
                'help': 'What folder / file do you want to copy over to the cluster',
                'action': 'store'
            }),
            (['-d', '--dest'], {
                'help': 'destination directory or file on the cluster',
                'action': 'store',
                'default': '/home/hadoop'
            }),
            (['-k', '--ssh-key'], {
                'help': 'location of ssh key to log into cluster nodes',
                'action': 'store',
                'required': True
            }),
            (['-p', '--profile'], {
                'help': 'AWS profile',
                'action': 'store',
                'default': 'default'
            })
        ]
    )
    def copy(self):
        cluster_id = self.app.pargs.cluster_id
        profile = self.app.pargs.profile
        ssh_key = self.app.pargs.ssh_key
        source = self.app.pargs.source
        dest = self.app.pargs.dest
        log = self.app.log

        session = boto3.session.Session(profile_name=profile)
        client = session.client('emr')
        paginator = client.get_paginator('list_instances')

        response_iterator = paginator.paginate(
            ClusterId=cluster_id
        )
        instances = []
        for response in response_iterator:
            for instance in response['Instances']:
                instances.append(ClusterInstance(
                    instance, {"key_filename": [ssh_key]}))

        for instance in instances:
            log.info("\n\nCopy file or folder %s, to %s:%s" %
                     (source, instance.public_dns_name, dest))
            result = instance.syncfiles(source, dest)
            log.info(str(result))

    @ex(
        help='boot a cluster based on the cluster config',
        arguments=[
            (['-p', '--profile'], {
                'help': 'AWS profile (default: default)',
                'action': 'store',
                'default': 'default'
            }),
            (['-n', '--name'], {
                'help': 'name of the keypair to generate',
                'action': 'store',
                'required': True
            }),
            (['-d', '--dir'], {
                'help': 'directory to put key',
                'action': 'store',
                'default': os.path.join(fs.HOME_DIR, '.ssh')
            })
        ]
    )
    def create_ec2_key(self):
        profile = self.app.pargs.profile
        name = self.app.pargs.name
        directory = self.app.pargs.dir

        session = boto3.session.Session(profile_name=profile)
        client = session.client('ec2')
        names = [pair['KeyName']
                 for pair in client.describe_key_pairs().get('KeyPairs', [])]

        if name in names:
            self.app.log.fatal(
                "Key already exists please specify a different key name")
            self.app.log.info(
                "These are the keys associated with your profile")
            self.app.log.info(yaml.dump(names, default_flow_style=False))
            return

        keypair = client.create_key_pair(KeyName=name)
        fs.ensure_dir_exists(directory)

        key_location = os.path.join(
            directory, "{name}".format(name=keypair['KeyName']))

        with open(key_location, 'w') as text_file:
            text_file.write(keypair['KeyMaterial'])

        self.app.log.info("Wrote new key to %s" % (key_location))
        os.chmod(key_location, stat.S_IRUSR | stat.S_IWUSR)
        self.app.log.info(
            "Changed permissions to prevent other users from reading the key")

    @ex(
        help='boot a cluster based on the cluster config',
        arguments=[
            (['-p', '--profile'], {
                'help': 'AWS profile (default: default)',
                'action': 'store',
                'default': 'default'
            }),
            (['-c', '--config'], {
                'help': 'Location of cluster config on the file system',
                'action': 'store',
                'required': True
            }),
            (['-k', '--ec2-key-name'], {
                'help': 'The ssh key, required if you want to log into the machine',
                'action': 'store'
            }),
            (['-Y'], {
                'help': 'Defaults will be chosen, useful for production',
                'action': 'store_true'
            }),
        ]
    )
    def boot(self):

        ec2_key_name = self.app.pargs.ec2_key_name or None
        profile = self.app.pargs.profile
        log = self.app.log
        config_path = self.app.pargs.config
        should_prompt = not self.app.pargs.Y

        if not ec2_key_name:
            log.warning("""
            You'll likely want to be able to log in or change dependencies of this cluster. 
            Without a key that will not be possible. If this is a dev cluster, it's almost a 
            certainty that you will want to specify a key.
            
            To create a key run: 
            
                cluster_funk clusters create-ec2-key -n my-key-name

            """)
            if should_prompt:
                resp = shell.Prompt("Do you still want to create the cluster?", options=[
                                    'yes', 'no'], numbered=True)
                if resp.input == 'no':
                    log.info("({input}) selected, please try again once you have create a ec2_key".format(
                        input=resp.input))
                    return

        try:
            table = self.app.db.table('users')
            user_id = table.all()[0]['id']
        except IndexError:
            user_id = None

        if not os.path.exists(config_path):
            log.info(
                "Could not find the cluster configuration you entered.  Please specify a valid path")
            return

        config = yaml.load(open(config_path).read())
        # TODO: create a config validator maybe to help with config issues.
        config['Tags'] = config.get('Tags', {})
        config['Tags']['user_id'] = user_id
        config['Ec2KeyName'] = self.app.pargs.ec2_key_name or None

        session = boto3.session.Session(profile_name=profile)

        booter = ClusterBooter(config, session)
        wait, resp = booter.boot()
        if wait:
            log.info(
                "Waiting for cluster to boot.  This can take about 10 minutes.  Feel free to grab a cup of ☕")
            wait()
            log.info("All set, cluster is fully booted.")

        table = self.app.db.table('history')
        table.insert({
            'action': 'cluster_boot',
            'resp': resp,
            'on': str(datetime.now())
        })

    @ex(
        help='given some stack information along with a job name, generate a config file',
        arguments=[
            (['-s', '--stack-prefix'], {
                'help': 'The cloudforamtion stack prefix (default: emr-spark-dev)',
                'action': 'store',
                'default': 'emr-spark'
            }),
            (['-p', '--profile'], {
                'help': 'AWS profile (default: default)',
                'action': 'store',
                'default': 'default'
            }),
            (['-j', '--job'], {
                'help': 'What job is this configuration for?',
                'action': 'store',
                'choices': job_names()
            }),
            (['-f', '--force'], {
                'help': 'Overwrite existing config if there is one',
                'action': 'store_true'
            }),
            (['-b', '--bucket'], {
                'help': 'S3 bucket used for logging and code storage',
                'action': 'store'
            }),
            (['-e', '--env'], {
                'help': 'Environment must be set',
                'action': 'store',
                'required': True,
                'choices': environment_names()
            })
        ]
    )
    def generate_config(self):
        args = self.app.pargs
        job = args.job or "unknown job"
        env = args.env or "dev"

        session = boto3.session.Session(profile_name=args.profile)
        client = session.client('cloudformation')
        stacks = StackCollection(client). \
            filter_by(StackCollection.is_cf_stack). \
            filter_by(partial(StackCollection.has_env, args.env)). \
            filter_by(partial(StackCollection.has_prefix, args.stack_prefix))

        output = stacks.output_dict()
        cluster_name = "Apache Spark for job {env}/{job}".format(
            env=env, job=job)
        # acquire latest Base AMI
        config = {
            'Name': cluster_name,
            'WorkLoadBucket': args.bucket,
            'WorkLoadPrefix': "apache_spark_for_job_{env}_{job}".format(env=env, job=job),
            'Schedule': {
                'Period': 'hourly',
                'Starting': datetime.now(),
                'Ending': datetime.now()
            },
            'Master': {
                'Instance': {
                    'Type': 'm4.large',
                    'Storage': 32
                }
            },
            'Worker': {
                'Instance': {
                    'Type': 'm4.large',
                    'Count': 1,
                    'Storage': 100,
                    'SpotCount': 0
                }
            },
            'KeepJobFlowAliveWhenNoSteps': True,
            'PipInstall': ["pandas==0.24.0"],
            'JobFlowRole': output.get('EMREc2RoleName', 'EMR_EC2_DefaultRole'),
            'ServiceRole': output.get('EMRRoleName', 'EMR_DefaultRole'),
            'AutoScalingRole': output.get('EMRAutoScalingRoleName', 'EMR_AutoScaling_DefaultRole'),
            'Ec2SubnetId': output.get('PublicSubnets', 'WARNING_INVALID_VALUE'),
            'EmrManagedMasterSecurityGroup': output.get('EmrManagedMasterSecurityGroup', 'WARNING_INVALID_VALUE'),
            'EmrManagedSlaveSecurityGroup': output.get('EmrManagedSlaveSecurityGroup', 'WARNING_INVALID_VALUE'),
            'EbsRootVolumeSize': 40,
            'AwsXrayEnabled': False,
            'ShutdownAfterInactivity': '2 hours',
            'Tags': {
                'Name': cluster_name,
                'Application': job,
                'job': job,
                'tool': 'cluster_funk',
                'Environment': env,
                'environment': env,
            }
        }

        # make sure env and job are valid!
        if env and job:
            root = find_nearest_root()
            env_config_path = os.path.join(
                root, 'jobs', job, 'config', 'clusters', env)
            config_path = os.path.join(env_config_path, "config.yml")
            fs.ensure_dir_exists(env_config_path)

            if not os.path.exists(config_path) or args.force:
                with open(config_path, "w") as text_file:
                    text_file.write(
                        yaml.dump(config, default_flow_style=False))
                self.app.log.info("Wrote a new config file too. {config_path}".format(
                    config_path=config_path))
            else:
                self.app.log.info(
                    "config already exists use -f if you really want to overwrite the existing one")

        self.app.log.info("\n%s" % yaml.dump(config, default_flow_style=False))

    @ex(
        help="Destroy the cluster given an id",
        arguments=[cluster_id_arg(), profile_arg()]
    )
    def destroy(self):
        profile = self.app.pargs.profile
        session = boto3.session.Session(profile_name=profile)
        client = session.client('emr')
        resp = client.terminate_job_flows(
            JobFlowIds=[self.app.pargs.cluster_id])
        self.app.log.info(resp)



import os

from cement import Controller, ex
from cement.utils import fs
from cement.utils.version import get_version_banner
from ..core.version import get_version

VERSION_BANNER = """
cluster_funk - a CLI for developing big data jobs %s
%s
""" % (get_version(), get_version_banner())


class Base(Controller):
    class Meta:
        label = 'base'

        # text displayed at the top of --help output
        description = 'Cli for managing EMR clusters for big data'

        # text displayed at the bottom of --help output
        epilog = 'Usage: cluster_funk new my_project_name'

        # controller level arguments. ex: 'cluster_funk --version'
        arguments = [
            # add a version banner
            (['-v', '--version'],
             {'action': 'version',
                'version': VERSION_BANNER}),
        ]

    def _default(self):
        """Default action if no sub-command is passed."""
        self.app.args.print_help()

    @ex(
        help='Create a new cluster_funk project',
        arguments=[
            (['name'], {'help': "what's the name of the project"})
        ]
    )
    def new(self):
        """Initialize EMR big data job project"""
        src = os.path.join(fs.abspath(__file__), os.pardir,
                           os.pardir, "templates", "base")
        dest = os.path.join(os.curdir, self.app.pargs.name)
        self.app.template.copy(src, dest, {})



import boto3


class StackDeployer:

    def __init__(self, user_id, profile, stack_prefix, env, stack_type, capabilities=['CAPABILITY_NAMED_IAM']):
        self.profile = profile
        self.capabilities = capabilities
        self.stack_prefix = stack_prefix
        self.env = env
        self.stack_type = stack_type
        self.user_id = user_id
        self.stack_name = "{stack_prefix}-{env}-{stack_type}".format(
            stack_prefix=stack_prefix, env=env, stack_type=stack_type)
        self.template_body = open(
            "./environments/{env}/{stack_type}.yml".format(env=env, stack_type=stack_type), 'r').read()

    def deploy(self):
        session = boto3.session.Session(profile_name=self.profile)
        client = session.client('cloudformation')
        try:
            response = client.create_stack(
                StackName=self.stack_name,
                TemplateBody=self.template_body,
                Capabilities=self.capabilities,
                Tags=[
                    {
                        'Key': 'Name',
                        'Value': self.stack_name
                    },
                    {
                        'Key': 'Application',
                        'Value': 'Emr Spark'
                    },
                    {
                        'Key': 'Tier',
                        'Value': 'APP'
                    },
                    {
                        'Key': 'Environment',
                        'Value': self.env.upper()
                    },
                    {
                        'Key': 'TechnicalTeam',
                        'Value': 'Architecture'
                    },
                    {
                        'Key': 'tool',
                        'Value': 'cluster_funk'
                    },
                    {
                        'Key': 'type',
                        'Value': self.stack_type
                    },
                    {
                        'Key': 'user_id',
                        'Value': self.user_id
                    },
                    {
                        'Key': 'environment',
                        'Value': self.env
                    }
                ],
                OnFailure='ROLLBACK',
                EnableTerminationProtection=True
            )
            waiter = client.get_waiter('stack_create_complete')

            def wait_func():
                waiter.wait(
                    StackName=response['StackId'],
                    WaiterConfig={
                        'Delay': 30,
                        'MaxAttempts': 60
                    }
                )
            return (wait_func, response)
        except client.exceptions.AlreadyExistsException:
            try:
                response = client.update_stack(
                    StackName=self.stack_name,
                    TemplateBody=self.template_body,
                    Capabilities=self.capabilities
                )
                waiter = client.get_waiter('stack_update_complete')

                def wait_func():
                    waiter.wait(
                        StackName=response['StackId'],
                        WaiterConfig={
                            'Delay': 30,
                            'MaxAttempts': 60
                        }
                    )
                return (wait_func, response)
            except client.exceptions.ClientError as nochanges:
                return (None, {'message': str("%s -> %s" % ("no changes to apply", nochanges))})

        return (None, {'message': 'no operation performed'})



import boto3
import yaml
from functools import partial

from cement import Controller, ex
from ..core.environments.stack_deployer import StackDeployer
from ..core.environments.stack_collection import StackCollection
from ..core.utils import environment_names, template_names


class Environments(Controller):

    class Meta:
        label = 'environments'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'Deploy and inspect your environment (security settings)'

    @ex(
        help='deploy environment',
        arguments=[
            (['-e', '--env'], {
                'help': 'Environment name (default: dev)',
                'action': 'store',
                'choices': environment_names(),
                'default': 'dev'
            }),
            (['-s', '--stack-prefix'], {
                'help': 'The cloudforamtion stack prefix (default: emr-spark)',
                'action': 'store',
                'default': 'emr-spark'
            }),
            (['-p', '--profile'], {
                'help': 'AWS profile (default: default)',
                'action': 'store',
                'default': 'default'
            })
        ]
    )
    def deploy(self):
        cliargs = self.app.pargs
        stack_prefix = cliargs.stack_prefix
        env = cliargs.env

        try:
            table = self.app.db.table('users')
            user_id = table.all()[0]['id']
        except IndexError:
            user_id = None

        if not user_id:
            self.app.log.error(
                "You must set your user_id in order to create a stack, for tagging :)")
            return None

        for stack_type in template_names(env):
            self.app.log.info("%s - Update Starting" % stack_type)

            deployer = StackDeployer(
                user_id=user_id,
                profile=cliargs.profile,
                stack_prefix=stack_prefix,
                env=env,
                stack_type=stack_type)

            wait, resp = deployer.deploy()

            if wait:
                self.app.log.info(
                    "Updating %s This can take a couple minutes please sit tight." % (stack_type))
                wait()

            self.app.log.info("%s - Update Complete" % stack_type)
            self.app.log.info(resp)

    @ex(
        help='Get outputs from cloudformation',
        arguments=[
            (['-e', '--env'], {
                'help': 'Environment name (default: dev)',
                'action': 'store',
                'choices': environment_names(),
                'default': 'dev'
            }),
            (['-p', '--profile'], {
                'help': 'AWS profile (default: default)',
                'action': 'store',
                'default': 'default'
            })
        ]
    )
    def outputs(self):
        profile_name = self.app.pargs.profile
        env = self.app.pargs.env
        for stack in self._get_cf_details(profile_name, env):
            self.app.log.info("Stack Id: {id}".format(id=stack['StackId']))
            print(yaml.dump(stack['Outputs']))

    @ex(
        help='Get full details from cloudformation',
        arguments=[
            (['-e', '--env'], {
                'help': 'Environment name (default: dev)',
                'action': 'store',
                'choices': environment_names(),
                'default': 'dev'
            }),
            (['-p', '--profile'], {
                'help': 'AWS profile (default: default)',
                'action': 'store',
                'default': 'default'
            })
        ]
    )
    def details(self):
        profile_name = self.app.pargs.profile
        env = self.app.pargs.env
        for stack in self._get_cf_details(profile_name, env):
            self.app.log.info("Stack Id: {id}".format(id=stack['StackId']))
            print(yaml.dump(stack))

    def _get_cf_details(self, profile_name, env):
        session = boto3.session.Session(profile_name=profile_name)
        client = session.client('cloudformation')
        stacks = StackCollection(client)
        stacks = stacks.filter_by(StackCollection.is_cf_stack)
        return stacks.filter_by(partial(StackCollection.has_env, env))



import logging

from .cluster_instance import ClusterInstance


class ClusterInstanceCollection:

    def __init__(self, client=None, cluster_id=None,
                 states=None, data=None, cluster_key=None, log=logging.Logger("cluster_funk")):
        self.log = log
        super()
        if data is None:
            self.list = []
            paginator = client.get_paginator('list_instances')
            response_iterator = paginator.paginate(ClusterId=cluster_id)

            for response in response_iterator:
                for instance in response['Instances']:
                    self.list.append(ClusterInstance(
                        instance, {"key_filename": [cluster_key]}))
        else:
            self.list = list(data)

    def __len__(self):
        return len(self.list)

    def __getitem__(self, ii):
        return self.list[ii]

    def __delitem__(self, ii):
        del self.list[ii]

    def __setitem__(self, ii, val):
        self.list[ii] = val

    def __str__(self):
        return str(self.list)

    def reverse(self):
        return self[::-1]

    def insert(self, ii, val):
        self.list.insert(ii, val)

    def filter_by(self, func):
        filtered = [instance for instance in self.list if func(instance)]
        return ClusterInstanceCollection(data=filtered)

    def syncfiles(self, src, dest):
        for ci in self.list:
            self.log.info("\n\nCopy file or folder %s, to %s:%s" %
                          (src, ci.public_dns_name, dest))
            result = ci.syncfiles(src, dest)
            self.log.info("Copied files:\n %s" % result)
        return self





class SagemakerNotebookBooter:
    pass

"""
    def __init__(self, cluster_id, session):
        self.cluster_id = cluster_id
        self.emr_client = session.client('emr') 
        self.details = None
        self.set_cluster_details()

    def set_cluster_details(self):
        if self.details:
            return self.details
        else:
            self.details = self.emr_client.describe_cluster(ClusterId=self.cluster_id)
            return self.details


    def _spark_magic_config(self, livy_url="http://localhost:8998"):
        return {
            "kernel_python_credentials": {
                "username": "",
                "password": "",
                "url": livy_url,
                "auth": "None"
            },
            "kernel_scala_credentials": {
                "username": "",
                "password": "",
                "url": livy_url,
                "auth": "None"
            },
            "kernel_r_credentials": {
                "username": "",
                "password": "",
                "url": livy_url
            },
            "logging_config": {
                "version": 1,
                "formatters": {
                    "magicsFormatter": {
                        "format": "%(asctime)s\t%(levelname)s\t%(message)s",
                        "datefmt": ""
                    }
                },
                "handlers": {
                    "magicsHandler": { 
                        "class": "hdijupyterutils.filehandler.MagicsFileHandler",
                        "formatter": "magicsFormatter",
                        "home_path": "~/.sparkmagic"
                    }
                },
                "loggers": {
                    "magicsLogger": {
                        "handlers": ["magicsHandler"],
                        "level": "DEBUG",
                        "propagate": 0
                    }
                }
            },
            "wait_for_idle_timeout_seconds": 30,
            "livy_session_startup_timeout_seconds": 120,
            "fatal_error_suggestion": "The code failed because of a fatal error:\n\t{}.\n\nSome things to try:\na) Make sure Spark has enough available resources for Jupyter to create a Spark context.\nb) Contact your Jupyter administrator to make sure the Spark magics library is configured correctly.\nc) Restart the kernel.",    
            "ignore_ssl_errors": False,
            "session_configs": {
                "driverMemory": "8000M",
                "executorCores": 4
            },
            "use_auto_viz": True,
            "coerce_dataframe": True,
            "max_results_sql": 2500,
            "pyspark_dataframe_encoding": "utf-8",
            "heartbeat_refresh_seconds": 30,
            "livy_server_heartbeat_timeout_seconds": 0,
            "heartbeat_retry_seconds": 10,
            "server_extension_default_kernel_name": "pysparkkernel",
            "custom_headers": {},
            "retry_policy": "configurable",
            "retry_seconds_to_sleep_list": [0.2, 0.5, 1, 3, 5],
            "configurable_retry_policy_max_retries": 8
        }

    def on_create(self):
        script=[
            '#!/bin/bash',
            'set -e',
            'FILE="/home/ec2-user/.sparkmagic/config.json"',
            'sudo /bin/cat <<EOM >$FILE',
            json.dump(self._spark_magic_config(), Indent=4)
            'EOM'
        ]
        return "\n".join(script)

    def on_start(self):
        return ""
    
    def boot(self):
        pass
"""



import uuid
from cement import Controller, ex


class UserId(Controller):

    class Meta:
        label = 'user_id'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = "Set a user_id to associated AWS resources with you"

    @ex(help='Set user id')
    def set(self):
        user_info = {
            'id': str(uuid.uuid4())
        }
        table = self.app.db.table('users')
        allusers = table.all()
        if not len(allusers):
            table.insert(user_info)
        else:
            user_info = allusers[0]
        self.app.log.info("Your userid has been set too %s" %
                          (user_info['id']))
        self.app.log.info(
            "Every resource you use will be tagged with this user_id")

    @ex(help='get user id')
    def get(self):
        table = self.app.db.table('users')
        try:
            self.app.log.info("Your user_id is: %s" % (table.all()[0]['id']))
        except IndexError:
            self.app.log.info(
                "No user_id set so can't get it. Please run user-id set first")

    @ex(help='clear user id')
    def clear(self):
        table = self.app.db.table('users')
        table.purge()



import hashlib


class ClusterBooter:

    def __init__(self, config, session):
        self.user_config = config
        self.emr_client = session.client('emr')
        self.s3_client = session.client('s3')
        self.bucket = config["WorkLoadBucket"]
        self.prefix = config["WorkLoadPrefix"]
        self._set_config()

    def _args_to_dict(self, **kwargs):
        return kwargs

    def _base_config(self):
        return self._args_to_dict(
            Name='placeholder',
            LogUri="placeholder",
            ReleaseLabel='emr-5.20.0',
            Instances={
                'InstanceGroups': [
                    {
                        'Name': 'MasterEMR',
                        'Market': 'ON_DEMAND',
                        'InstanceRole': 'MASTER',
                        'InstanceType': 'm1.medium',
                        'InstanceCount': 1,
                        'EbsConfiguration': {
                            'EbsBlockDeviceConfigs': [{
                                'VolumeSpecification': {
                                    'SizeInGB': 32,
                                    'VolumeType': 'gp2'
                                },
                                'VolumesPerInstance': 1
                            }],
                            'EbsOptimized': False
                        }
                    },
                    {
                        'Name': 'WorkerEMR',
                        'Market': 'ON_DEMAND',
                        'InstanceCount': 1,
                        'InstanceType': 'm1.medium',
                        'InstanceRole': 'CORE',
                        'EbsConfiguration': {
                            'EbsBlockDeviceConfigs': [{
                                'VolumeSpecification': {
                                    'SizeInGB': 100,
                                    'VolumeType': 'gp2'
                                },
                                'VolumesPerInstance': 1
                            }],
                            'EbsOptimized': False
                        }
                    }
                ],
                # 'Ec2KeyName': 'Dcos',
                'KeepJobFlowAliveWhenNoSteps': True,
                'TerminationProtected': False,
                'Ec2SubnetId': 'placeholder',
                'EmrManagedMasterSecurityGroup': 'placeholder',
                'EmrManagedSlaveSecurityGroup': 'placeholder'
            },
            Steps=[],
            BootstrapActions=[],
            Applications=[
                {
                    'Name': 'Spark'
                },
                {
                    'Name': 'Livy'
                },
                {
                    'Name': 'Hadoop'
                },
                {
                    'Name': 'Hive'
                }
            ],
            Configurations=[
                {
                    'Classification': 'emrfs-site',
                    'Properties': {
                        "fs.s3.consistent.retryPeriodSeconds": "10",
                        "fs.s3.consistent": "true",
                        "fs.s3.consistent.retryCount": "5",
                        "fs.s3.consistent.metadata.tableName": "EmrFSMetadata"
                    }
                },
                {
                    "Classification": "spark-defaults",
                    "Properties": {
                        "spark.yarn.appMasterEnv.SPARK_HOME": "/usr/lib/spark",
                        "spark.yarn.appMasterEnv.PYSPARK_PYTHON": "/usr/bin/python3"
                    }
                },
                {
                    "Classification": "yarn-site",
                    "Properties": {
                        "yarn.nodemanager.vmem-check-enabled": "false"
                    }
                },
                {
                    "Classification": "spark-env",
                    "Configurations": [
                        {
                            "Classification": "export",
                            "Properties": {
                                "PYSPARK_PYTHON": "/usr/bin/python3"
                            }
                        }
                    ]
                },
                {
                    "Classification": "spark-hive-site",
                    "Properties": {
                        "hive.metastore.client.factory.class": "com.amazonaws.glue.catalog.metastore.AWSGlueDataCatalogHiveClientFactory"
                    }
                }
            ],
            VisibleToAllUsers=True,
            JobFlowRole='place_holder',
            ServiceRole='place_holder',
            AutoScalingRole='place_holder',
            ScaleDownBehavior='TERMINATE_AT_TASK_COMPLETION',
            EbsRootVolumeSize=40
        )

    def _set_config(self):
        # the task is to mix these two configs together to get a final config.
        base = self._base_config().copy()
        user_config = self.user_config.copy()

        cluster_properties = (
            'JobFlowRole',
            'ServiceRole',
            'AutoScalingRole',
            'Name',
            'EbsRootVolumeSize'
        )

        for item in cluster_properties:
            user_config_item = user_config.get(item, None)
            if user_config_item:
                base[item] = user_config_item
                user_config_item = None

        if user_config.get('Ami', None):
            base['CustomAmiId'] = user_config['Ami']

        instances = base['Instances']
        instance_properties = (
            'KeepJobFlowAliveWhenNoSteps',
            'Ec2SubnetId',
            'EmrManagedMasterSecurityGroup',
            'EmrManagedSlaveSecurityGroup'
        )
        for item in instance_properties:
            user_config_item = user_config.get(item, None)
            if user_config_item:
                instances[item] = user_config_item
                user_config_item = None

        base['Tags'] = base.get('Tags', [])
        for k, v in user_config.get('Tags', {}).items():
            base['Tags'].append({'Key': k, 'Value': v})

        for group in base['Instances']['InstanceGroups']:
            if group['InstanceRole'] == 'MASTER':
                group['InstanceType'] = user_config['Master']['Instance']['Type']
                for device_config in group['EbsConfiguration']['EbsBlockDeviceConfigs']:
                    device_config['VolumeSpecification']['SizeInGB'] = user_config['Master']['Instance']['Storage']

            if group['InstanceRole'] == 'CORE':
                group['InstanceCount'] = user_config['Worker']['Instance']['Count']
                group['InstanceType'] = user_config['Worker']['Instance']['Type']
                for device_config in group['EbsConfiguration']['EbsBlockDeviceConfigs']:
                    device_config['VolumeSpecification']['SizeInGB'] = user_config['Worker']['Instance']['Storage']

        base['LogUri'] = "s3n://{bucket}/{prefix}/logs".format(
            bucket=user_config["WorkLoadBucket"],
            prefix=user_config["WorkLoadPrefix"]
        )
        if user_config['PipInstall']:
            script = self._generate_pip_script(user_config['PipInstall'])
            location = self._save_script_to_s3(script)
            base['BootstrapActions'].append({
                'Name': 'Install Python Libs',
                'ScriptBootstrapAction': {
                    'Path': location
                }
            })

        if user_config['AwsXrayEnabled']:
            script = self._generate_xray_script()
            location = self._save_script_to_s3(script)
            base['BootstrapActions'].append({
                'Name': 'Install AWS Xray',
                'ScriptBootstrapAction': {
                    'Path': location
                }
            })

        if user_config['Ec2KeyName']:
            base['Instances']['Ec2KeyName'] = user_config['Ec2KeyName']

        self.config = base

        return self.config

    def _generate_pip_script(self, pips):
        script = [
            '#!/bin/bash',
            'set -e',
            'FILE="/home/hadoop/requirements.txt"',
            'sudo /bin/cat <<EOM >$FILE'
        ]

        for pip in pips:
            script.append(pip)

        script.append('EOM')
        script.append('sudo /usr/bin/python3 -m pip install -r $FILE')
        return "\n".join(script)

    def _generate_xray_script(self):
        script = [
            '#!/bin/bash',
            'set -e',
            'curl https://s3.dualstack.us-east-1.amazonaws.com/aws-xray-assets.us-east-1/xray-daemon/aws-xray-daemon-3.x.rpm -o /home/hadoop/xray.rpm',
            'sudo yum install -y /home/hadoop/xray.rpm'
        ]
        return "\n".join(script)

    def _save_script_to_s3(self, script):
        script_bytes = script.encode('utf-8')
        file_name = hashlib.md5(script_bytes).hexdigest()
        self.s3_client.put_object(
            Body=script_bytes,
            Bucket=self.bucket,
            Key="%s/scripts/%s.sh" % (self.prefix, file_name)
        )
        return "s3://%s/%s/scripts/%s.sh" % (self.bucket, self.prefix, file_name)

    def boot(self):
        response = self.emr_client.run_job_flow(**self.config)
        waiter = self.emr_client.get_waiter('cluster_running')

        def wait():
            waiter.wait(
                ClusterId=response['JobFlowId'],
                WaiterConfig={
                    'Delay': 30,
                    'MaxAttempts': 30
                }
            )
        return wait, response



class JobCollection:

    def __init__(
            self, client=None, cluster_id=None,
            job_id=None, states=None, data=None):

        super()

        if data is None:
            self.list = list()
            if job_id:
                step = client.describe_step(
                    ClusterId=cluster_id,
                    StepId=job_id
                )
                self.list.append(step['Step'])
            else:
                pager_args = {
                    'ClusterId': cluster_id
                }

                if states:
                    pager_args['StepStates'] = states

                paginator = client.get_paginator('list_steps')
                results = paginator.paginate(**pager_args)
                for result in results:
                    for step in result['Steps']:
                        self.list.append(step)
        else:
            self.list = list(data)

    def __len__(self):
        return len(self.list)

    def __getitem__(self, ii):
        return self.list[ii]

    def __delitem__(self, ii):
        del self.list[ii]

    def __setitem__(self, ii, val):
        self.list[ii] = val

    def __str__(self):
        return str(self.list)

    def reverse(self):
        return self[::-1]

    def insert(self, ii, val):
        self.list.insert(ii, val)

    def filter_by(self, func):
        filtered = [job for job in self.list if func(job)]
        return JobCollection(data=filtered)



import os
import uuid
import tarfile

from fabric.connection import Connection
from cement.utils import fs


class ClusterInstance:

    # pass the connection class in just in case we want to send in a mock
    def __init__(self, instance_params, connect_kwargs={}):
        self.intance_params = instance_params
        self.id = instance_params['Id']
        self.public_dns_name = instance_params['PublicDnsName']
        self.public_ip = instance_params['PublicIpAddress']
        self.state = instance_params['Status']['State']
        self.running = False

        if self.state == 'RUNNING':
            self.running = True

        self.connection = Connection(
            host=self.public_dns_name,
            user='hadoop',
            port=22,
            connect_kwargs=connect_kwargs
        )

    def run_cmd(self, cmd):
        return self.connection.run(cmd)

    def sudo_cmd(self, cmd):
        return self.connection.sudo(cmd)

    def syncfiles(self, src, dest):

        if not self.running:
            raise "That Cluster Instance isn't running, can't copy files"

        if os.path.isdir(src):

            EXCLUDE_FILES = ['__pycache__', '.git', '.gitignore']

            def filter_function(tarinfo):
                for filename in EXCLUDE_FILES:
                    if filename in tarinfo.name:
                        return None
                return tarinfo

            with fs.Tmp() as tmp:

                filename = uuid.uuid4()
                temp_file = "%s/%s.tar.gz" % (tmp.dir, filename)

                with tarfile.open(temp_file, "w:gz") as tar:
                    tar.add(src, recursive=True, arcname=str(
                        filename), filter=filter_function)

                self.connection.put(temp_file, remote='/mnt/tmp')
                self.run_cmd(
                    "tar -C /mnt/tmp -xzf /mnt/tmp/%s.tar.gz" % (filename))
                self.run_cmd("rm -rf %s" % (dest))
                return self.run_cmd("mv /mnt/tmp/%s %s" % (filename, dest))

        if os.path.isfile(src):
            return self.connection.put(src, remote=dest)



class StackCollection:

    def __init__(self, client=None, data=None):
        super(StackCollection, self).__init__()
        if data == None:
            paginator = client.get_paginator('describe_stacks')
            results = paginator.paginate()
            self.list = list()
            for result in results:
                stacks = result['Stacks']
                for stack in stacks:
                    self.list.append(stack)
        else:
            self.list = list(data)

    def __len__(self):
        return len(self.list)

    def __getitem__(self, ii):
        return self.list[ii]

    def __delitem__(self, ii):
        del self.list[ii]

    def __setitem__(self, ii, val):
        self.list[ii] = val

    def __str__(self):
        return str(self.list)

    def insert(self, ii, val):
        self.list.insert(ii, val)

    def filter_by(self, func):
        filtered = [stack for stack in self.list if func(stack)]
        return StackCollection(data=filtered)

    @staticmethod
    def has_prefix(stack_prefix, stack):
        for tag in stack.get('Tags', []):
            if tag['Key'] == 'Name' and tag.get('Value', "").startswith(stack_prefix):
                return True
        return False

    @staticmethod
    def is_cf_stack(stack):
        for tag in stack.get('Tags', []):
            if tag['Key'] == 'tool' and tag['Value'] == 'cluster_funk':
                return True
        return False

    @staticmethod
    def has_env(env, stack):
        for tag in stack.get('Tags', []):
            if tag['Key'] == 'environment' and tag['Value'] == env:
                return True
        return False

    def output_dict(self):
        result = {}

        for stack in self:
            for output in stack.get("Outputs", []):
                result[output.get("OutputKey", "")] = output["OutputValue"]

        return result



import re
import argparse


def cluster_id(overwrite={}):
    opts = {
        'help': 'Cluster id that the job ran on',
        'required': True
    }
    return (['-c', '--cluster-id'], {**opts, **overwrite})


def profile(overwrite={}):
    opts = {
        'help': 'AWS profile',
        'default': 'default'
    }
    return (['-p', '--profile'], {**opts, **overwrite})


def job_id(overwrite={}):
    opts = {
        'help': 'filter results by a specific job id'
    }
    return (['-j', '--job-id'], {**opts, **overwrite})


def job_state(overwrite={}):
    opts = {
        'help': 'Filter by state.  Can be used multiple times',
        'choices': ['PENDING', 'CANCEL_PENDING', 'RUNNING', 'COMPLETED', 'CANCELLED', 'FAILED', 'INTERRUPTED'],
        'action': 'append'
    }
    return (['-s', '--state'], {**opts, **overwrite})


def job_name(overwrite={}):

    def name_type(string):
        if not re.match(r'^\w+$', string):
            raise argparse.ArgumentTypeError
        return string

    opts = {
        'help': 'The name of the job valid format [A-Z][a-z][0-9]_',
        'action': 'store',
        'required': True,
        'type': name_type
    }

    return (['-n', '--name'], {**opts, **overwrite})


def job_on_failure(overwrite={}):
    opts = {
        'help': 'Action on failure of the job',
        'choices': ['TERMINATE_JOB_FLOW', 'TERMINATE_CLUSTER', 'CANCEL_AND_WAIT', 'CONTINUE'],
        'default': 'CONTINUE'
    }
    return (['-f', '--on-failure'], {**opts, **overwrite})


def job_on_success(overwrite={}):
    opts = {
        'help': 'Action on step success',
        'choices': ['TERMINATE_CLUSTER', 'CONTINUE'],
        'default': 'CONTINUE'
    }
    return (['-s', '--on-success'], {**opts, **overwrite})


def job_argument(overwrite={}):
    opts = {
        'help': 'Optional arguments to pass to your main.py use more than once for many args',
        'action': 'append',
        'type': str
    }
    return (['-a', '--argument'], {**opts, **overwrite})


def ssh_key_location(overwrite={}):
    opts = {
        'help': 'location of ssh key to log into cluster nodes',
        'required': True
    }
    return (['-k', '--ssh-key'], {**opts, **overwrite})



import os
import uuid
import shutil

from tinydb import TinyDB
from cement import App, TestApp, init_defaults
from cement.ext.ext_logging import LoggingLogHandler
from cement.core.exc import CaughtSignal
from cement.utils import fs
from cement.core.template import TemplateHandler
from jinja2 import Template
from .core.exc import ClusterFunkError
from .controllers.base import Base
from .controllers.clusters import Clusters
from .controllers.user_id import UserId
from .controllers.environments import Environments
from .controllers.jobs import Jobs

# configuration defaults
CONFIG = init_defaults('cluster_funk')
CONFIG['cluster_funk']['db_file'] = '~/.cluster_funk/db.json'


class CustomTemplateHandler(TemplateHandler):

    class Meta:
        label = "withcopy"

    def render(self, content, data):
        template = Template(content)
        return template.render(**data)


def extend_tinydb(app):
    app.log.info('extending cluster_funk application with tinydb')
    db_file = app.config.get('cluster_funk', 'db_file')

    db_file = fs.abspath(db_file)
    app.log.info('tinydb database file is: %s' % db_file)

    db_dir = os.path.dirname(db_file)
    if not os.path.exists(db_dir):
        os.makedirs(db_dir)

    db = TinyDB(db_file)

    user_info = {'id': str(uuid.uuid4())}
    table = db.table('users')
    allusers = table.all()
    if not len(allusers):
        table.insert(user_info)
    else:
        user_info = allusers[0]

    app.extend('db', db)

def rm_cache_from_templates(app):
    app.log.info('Ensure template directory is clean')
    dir_path = os.path.dirname(os.path.realpath(__file__))
    tp = os.path.join(dir_path, "templates")
    pycache_dirs = [d for d, s, f in os.walk(tp) if "__pycache__" in d]
    for directory in pycache_dirs:
        app.log.info(directory)
        app.log.info("removed %s" % directory)
        shutil.rmtree(directory)

class ClusterFunk(App):
    """Cluster Funk primary application."""

    class Meta:
        label = 'cluster_funk'

        hooks = [
            ('post_setup', extend_tinydb),
            ('post_setup', rm_cache_from_templates)
        ]

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        close_on_exit = True

        # load additional framework extensions
        extensions = [
            'yaml',
            'colorlog',
            'jinja2',
        ]

        # configuration handler
        config_handler = 'yaml'

        # configuration file suffix
        config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'jinja2'

        template_handler = CustomTemplateHandler

        # register handlers
        handlers = [
            Base,
            Clusters,
            UserId,
            Environments,
            Jobs
        ]


class ClusterFunkTestLoggingLogHandler(LoggingLogHandler):
    pass


class ClusterFunkTest(TestApp, ClusterFunk):
    """A sub-class of ClusterFunk that is better suited for testing."""

    class Meta:
        label = 'cluster_funk'
        log_handler = ClusterFunkTestLoggingLogHandler


def main():
    with ClusterFunk() as app:
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except ClusterFunkError as e:
            print('ClusterFunkError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()



import os


def find_nearest_root(current_path=os.getcwd(), iterleft=20):
    if not iterleft:
        return None
    if os.path.exists("%s/.cf" % (current_path)):
        return current_path
    return find_nearest_root(os.path.abspath(os.path.join(current_path, os.pardir)), iterleft=iterleft-1)


def get_subfolders(start_path):
    try:
        root = find_nearest_root()
        full_path = os.path.join(root, start_path)
        if os.path.exists(full_path):
            return os.listdir(full_path)
        return list()
    except TypeError:
        return list()


def environment_names():
    return get_subfolders("environments")


def template_names(env):
    return [filename.split(".")[0] for filename in get_subfolders("environments/%s" % (env))]


def job_names():
    return get_subfolders("jobs")



import logging


class EmrCluster:

    def __init__(self, client, cluster_id, log=logging.Logger("cluster_funk")):
        self.cluster_id = cluster_id
        self.client = client

    def submit_spark_job(self, job_path, arguments=[], on_failure='CONTINUE'):
        base_args = [
            'spark-submit',
            '--deploy-mode',
            'client',
            '%s/main.py' % (job_path)
        ]
        for argument in arguments:
            replace_argument = argument.replace('__JOB_PATH__', job_path)
            base_args.append(replace_argument)

        return self.client.add_job_flow_steps(
            JobFlowId=self.cluster_id,
            Steps=[
                {
                    'Name': "Spark cluster_funk",
                    'ActionOnFailure': on_failure,
                    'HadoopJarStep': {
                        'Jar': 'command-runner.jar',
                        'Args': base_args
                    }
                }
            ]
        )




from setuptools import setup, find_packages
from cluster_funk.core.version import get_version

VERSION = get_version()

f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='cluster_funk',
    version=VERSION,
    description='CLI for managing EMR clusters for big data',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Jeff Durand',
    author_email='jeff.durand@gmail.com',
    url='https://bitbucket.org/jeffdurand/cluster_funk',
    license='unlicensed',
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    package_data={'cluster_funk': ['templates/*']},
    include_package_data=True,
    install_requires=[
        "tinydb",
        "boto3",
        "colorlog",
        "jinja2",
        "pyyaml",
        "fabric"
    ],
    entry_points="""
        [console_scripts]
        cluster_funk = cluster_funk.main:main
    """,
)



import boto3
import uuid
import os
import yaml

from cement import Controller, ex
from cement.utils import fs

from ..core.utils import find_nearest_root
from ..core.jobs.job_collection import JobCollection
from ..core.clusters.cluster_instance_collection import ClusterInstanceCollection
from ..core.clusters.emr_cluster import EmrCluster

from .arguments import profile as profile_arg
from .arguments import job_id as job_id_arg
from .arguments import job_state as job_state_arg
from .arguments import cluster_id as cluster_id_arg
from .arguments import job_name as job_name_arg
from .arguments import job_argument as job_argument_arg
from .arguments import ssh_key_location as ssh_key_location_arg
from .arguments import job_on_failure as job_on_failure_arg

def emr_client(pargs):
    session = boto3.session.Session(profile_name=pargs.profile)
    return session.client('emr')


class Jobs(Controller):

    class Meta:
        label = 'jobs'
        stacked_type = 'nested'
        stacked_on = 'base'
        description = 'Commands for creating and submitting jobs to running clusters'
        epilog = "Usage: cluster_funk jobs generate -n my_great_new_job"
        help = "Create and submit jobs to a cluster"

    @ex(
        help='Get the most recent 100 jobs from a cluster',
        arguments=[
            cluster_id_arg(),
            job_id_arg(),
            profile_arg(),
            job_state_arg()
        ]
    )
    def list(self):
        # most_recent_cluster_id(self.app)
        client = emr_client(self.app.pargs)
        cluster_id = self.app.pargs.cluster_id
        job_id = self.app.pargs.job_id
        states = self.app.pargs.state

        jobs = JobCollection(
            client=client,
            cluster_id=cluster_id,
            job_id=job_id,
            states=states
        )

        if len(jobs):
            for job in jobs.reverse():
                self.app.log.info(
                    "\n----------------------------\n%s" % (yaml.dump(job)))
        else:
            self.app.log.info(
                "We didn't find any jobs for that cluster.  please check the cluster id")

    @ex(
        help='Bundle up a job and submit it to the running cluster',
        arguments=[
            job_name_arg(),
            cluster_id_arg(),
            profile_arg(),
            job_argument_arg(),
            ssh_key_location_arg(),
            job_on_failure_arg()
        ]
    )
    def submit(self):
        root_path = find_nearest_root()
        profile = self.app.pargs.profile
        cluster_id = self.app.pargs.cluster_id
        on_failure = self.app.pargs.on_failure
        name = self.app.pargs.name
        ssh_key = self.app.pargs.ssh_key
        arguments = self.app.pargs.argument
        log = self.app.log

        session = boto3.session.Session(profile_name=profile)
        client = session.client('emr')

        job_dir = "{root}/jobs/{name}".format(root=root_path, name=name)
        job_folder_id = str(uuid.uuid4())
        dest = "/mnt/tmp/%s" % (job_folder_id)

        log.info("Staging job run: %s" % (job_folder_id))

        instances = ClusterInstanceCollection(
            client=client,
            cluster_id=cluster_id,
            cluster_key=ssh_key,
            log=log
        )
        instances.syncfiles(job_dir, dest)

        log.info("Add folder to hdfs")
        instances[0].run_cmd("hadoop fs -mkdir -p /mnt/tmp")
        instances[0].run_cmd("hadoop fs -put %s %s" % (dest, dest))
        cluster = EmrCluster(
            client=client,
            cluster_id=cluster_id,
            log=log
        )
        job_params = {
            'job_path': dest,
            'on_failure': on_failure
        }
        if arguments:
            job_params['arguments'] = arguments

        response = cluster.submit_spark_job(**job_params)

        log.info(response)

    @ex(
        help='Generate a new job in the jobs directory',
        arguments=[
            job_name_arg()
        ]
    )
    def generate(self):
        root_path = find_nearest_root()

        if not root_path:
            self.app.log.fatal(
                "Could not find the root of the project please change directories")
            return None

        fs.ensure_dir_exists("%s/jobs" % root_path)
        job_dir = "{root}/jobs/{name}".format(
            root=root_path, name=self.app.pargs.name)

        if os.path.exists(job_dir):
            self.app.log.fatal(
                "Sorry there is already a job in this directory can't overwrite")
            return None

        example_path = os.path.join(fs.abspath(
            __file__), os.pardir, os.pardir, "templates", "word_count")
        data = {
            "name": self.app.pargs.name
        }
        self.app.template.copy(example_path, job_dir, data)
        self.app.log.info("""
            A cluster config is not created when you generate a job,
            please refer to the job README.md for instructions on how to generate
            a cluster config file that can be used to run jobs on EMR

            Alternatively you can type if you are familiar with the process:

            cluster_funk clusters generate-config -h

        """)
